@@ -2,7 +2,8 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2006-2007  Nokia Corporation
+ *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -59,6 +60,7 @@
 #include "headset.h"
 #include "gateway.h"
 #include "sink.h"
+#include "control.h"
 #include "manager.h"
 
 typedef enum {
@@ -105,7 +107,13 @@ static uint32_t hf_record_id = 0;
 static GIOChannel *hs_server = NULL;
 static GIOChannel *hf_server = NULL;
 
-static const struct enabled_interfaces *enabled;
+static struct enabled_interfaces enabled = {
+	.headset	= TRUE,
+	.gateway	= FALSE,
+	.sink		= TRUE,
+	.source		= FALSE,
+	.control	= TRUE,
+};
 
 static void get_next_record(struct audio_sdp_data *data);
 static DBusHandlerResult get_handles(const char *uuid,
@@ -211,7 +219,7 @@ done:
 	return uuid16;
 }
 
-static gboolean server_is_enabled(uint16_t svc)
+gboolean server_is_enabled(uint16_t svc)
 {
 	gboolean ret;
 
@@ -223,7 +231,10 @@ static gboolean server_is_enabled(uint16_t svc)
 		ret = (hf_server != NULL);
 		break;
 	case AUDIO_SINK_SVCLASS_ID:
-		return enabled->sink;
+		return enabled.sink;
+	case AV_REMOTE_TARGET_SVCLASS_ID:
+	case AV_REMOTE_SVCLASS_ID:
+		return enabled.control;
 	default:
 		ret = FALSE;
 		break;
@@ -275,9 +286,17 @@ static void handle_record(sdp_record_t *record, struct device *device)
 		break;
 	case AV_REMOTE_SVCLASS_ID:
 		debug("Found AV Remote");
+		if (device->control == NULL)
+			device->control = control_init(device);
+		if (device->sink && sink_is_active(device))
+			avrcp_connect(device);
 		break;
 	case AV_REMOTE_TARGET_SVCLASS_ID:
 		debug("Found AV Target");
+		if (device->control == NULL)
+			device->control = control_init(device);
+		if (device->sink && sink_is_active(device))
+			avrcp_connect(device);
 		break;
 	default:
 		debug("Unrecognized UUID: 0x%04X", uuid16);
@@ -314,7 +333,7 @@ static void finish_sdp(struct audio_sdp_data *data, gboolean success)
 	if (dbus_error_is_set(&derr)) {
 		error("Unable to get message args");
 		success = FALSE;
-		err_failed(connection, data->msg, derr.message);
+		error_failed(connection, data->msg, derr.message);
 		dbus_error_free(&derr);
 		goto done;
 	}
@@ -323,14 +342,14 @@ static void finish_sdp(struct audio_sdp_data *data, gboolean success)
 	if (!data->records) {
 		debug("No audio audio related service records were found");
 		success = FALSE;
-		err_not_supported(connection, data->msg);
+		error_not_supported(connection, data->msg);
 		goto done;
 	}
 
 	reply = dbus_message_new_method_return(data->msg);
 	if (!reply) {
 		success = FALSE;
-		err_failed(connection, data->msg, "Out of memory");
+		error_failed(connection, data->msg, "Out of memory");
 		goto done;
 	}
 
@@ -379,13 +398,14 @@ static void get_record_reply(DBusPendingCall *call,
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
+		/* FIXME : forward error message as is */
 		error("GetRemoteServiceRecord failed: %s", derr.message);
 		if (dbus_error_has_name(&derr,
 					"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connect_failed(connection, data->msg,
-				strerror(EHOSTDOWN));
+			error_connection_attempt_failed(connection, data->msg,
+				EHOSTDOWN);
 		else
-			err_failed(connection, data->msg, derr.message);
+			error_failed(connection, data->msg, derr.message);
 		dbus_error_free(&derr);
 		goto failed;
 	}
@@ -393,7 +413,7 @@ static void get_record_reply(DBusPendingCall *call,
 	if (!dbus_message_get_args(reply, NULL,
 				DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE, &array, &array_len,
 				DBUS_TYPE_INVALID)) {
-		err_failed(connection, data->msg,
+		error_failed(connection, data->msg,
 				"Unable to get args from GetRecordReply");
 		goto failed;
 	}
@@ -439,7 +459,7 @@ static void get_next_record(struct audio_sdp_data *data)
 						"GetRemoteServiceRecord");
 	if (!msg) {
 		error("Unable to allocate new method call");
-		err_connect_failed(connection, data->msg, strerror(ENOMEM));
+		error_out_of_memory(connection, data->msg);
 		finish_sdp(data, FALSE);
 		return;
 	}
@@ -458,7 +478,7 @@ static void get_next_record(struct audio_sdp_data *data)
 
 	if (!dbus_connection_send_with_reply(connection, msg, &pending, -1)) {
 		error("Sending GetRemoteServiceRecord failed");
-		err_connect_failed(connection, data->msg, strerror(EIO));
+		error_connection_attempt_failed(connection, data->msg, EIO);
 		finish_sdp(data, FALSE);
 		return;
 	}
@@ -493,12 +513,14 @@ static void get_handles_reply(DBusPendingCall *call,
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
+		/* FIXME : forward error message as is */
 		error("GetRemoteServiceHandles failed: %s", derr.message);
 		if (dbus_error_has_name(&derr,
 					"org.bluez.Error.ConnectionAttemptFailed"))
-			err_connect_failed(connection, data->msg, strerror(EHOSTDOWN));
+			error_connection_attempt_failed(connection, data->msg,
+							EHOSTDOWN);
 		else
-			err_failed(connection, data->msg, derr.message);
+			error_failed(connection, data->msg, derr.message);
 		dbus_error_free(&derr);
 		goto failed;
 	}
@@ -507,7 +529,7 @@ static void get_handles_reply(DBusPendingCall *call,
 				DBUS_TYPE_ARRAY, DBUS_TYPE_UINT32,
 				&array, &array_len,
 				DBUS_TYPE_INVALID)) {
-		err_failed(connection, data->msg,
+		error_failed(connection, data->msg,
 				"Unable to get args from reply");
 		goto failed;
 	}
@@ -558,7 +580,7 @@ static DBusHandlerResult get_handles(const char *uuid,
 						"org.bluez.Adapter",
 						"GetRemoteServiceHandles");
 	if (!msg) {
-		err_failed(connection, data->msg,
+		error_failed(connection, data->msg,
 				"Could not create a new dbus message");
 		goto failed;
 	}
@@ -570,7 +592,7 @@ static DBusHandlerResult get_handles(const char *uuid,
 					DBUS_TYPE_INVALID);
 
 	if (!dbus_connection_send_with_reply(connection, msg, &pending, -1)) {
-		err_failed(connection, data->msg,
+		error_failed(connection, data->msg,
 				"Sending GetRemoteServiceHandles failed");
 		goto failed;
 	}
@@ -625,7 +647,7 @@ struct device *manager_device_connected(bdaddr_t *bda, const char *uuid)
 		created = TRUE;
 	}
 
-	if (!strcmp(uuid, HSP_AG_UUID) || !strcmp(uuid, HSP_AG_UUID)) {
+	if (!strcmp(uuid, HSP_AG_UUID) || !strcmp(uuid, HFP_AG_UUID)) {
 		if (device->headset)
 			return device;
 
@@ -635,8 +657,7 @@ struct device *manager_device_connected(bdaddr_t *bda, const char *uuid)
 			return NULL;
 
 		headset = TRUE;
-	}
-	else if (!strcmp(uuid, A2DP_SOURCE_UUID)) {
+	} else if (!strcmp(uuid, A2DP_SOURCE_UUID)) {
 		if (device->sink)
 			return device;
 
@@ -644,6 +665,14 @@ struct device *manager_device_connected(bdaddr_t *bda, const char *uuid)
 
 		if (!device->sink)
 			return NULL;
+	} else if (!strcmp(uuid, AVRCP_TARGET_UUID)) {
+		if (device->control)
+			return device;
+
+		device->control = control_init(device);
+
+		if (!device->control)
+			return NULL;
 	} else
 		return NULL;
 
@@ -716,7 +745,7 @@ static DBusHandlerResult am_create_device(DBusConnection *conn,
 				DBUS_TYPE_INVALID);
 
 	if (dbus_error_is_set(&derr)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		dbus_error_free(&derr);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
@@ -806,18 +835,18 @@ static DBusHandlerResult am_remove_device(DBusConnection *conn,
 	if (!dbus_message_get_args(msg, &derr,
 					DBUS_TYPE_STRING, &path,
 					DBUS_TYPE_INVALID)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 	if (dbus_error_is_set(&derr)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		dbus_error_free(&derr);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 
 	match = g_slist_find_custom(devices, path, device_path_cmp);
 	if (!match)
-		return err_does_not_exist(connection, msg);
+		return error_device_does_not_exist(connection, msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -887,7 +916,7 @@ static DBusHandlerResult am_find_by_addr(DBusConnection *conn,
 				DBUS_TYPE_STRING, &address,
 				DBUS_TYPE_INVALID);
 	if (dbus_error_is_set(&derr)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		dbus_error_free(&derr);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
@@ -896,7 +925,7 @@ static DBusHandlerResult am_find_by_addr(DBusConnection *conn,
 	device = manager_find_device(&bda, NULL, FALSE);
 
 	if (!device)
-		return err_does_not_exist(conn, msg);
+		return error_device_does_not_exist(conn, msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -915,12 +944,12 @@ static DBusHandlerResult am_default_device(DBusConnection *conn,
 	DBusMessage *reply;
 
 	if (!default_dev)
-		return err_does_not_exist(connection, msg);
+		return error_device_does_not_exist(connection, msg);
 
 	if (default_dev->headset == NULL &&
 		dbus_message_is_method_call(msg, AUDIO_MANAGER_INTERFACE,
 							"DefaultHeadset"))
-		return err_does_not_exist(connection, msg);
+		return error_device_does_not_exist(connection, msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -946,18 +975,18 @@ static DBusHandlerResult am_change_default_device(DBusConnection *conn,
 	if (!dbus_message_get_args(msg, &derr,
 					DBUS_TYPE_STRING, &path,
 					DBUS_TYPE_INVALID)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 	if (dbus_error_is_set(&derr)) {
-		err_invalid_args(connection, msg, derr.message);
+		error_invalid_arguments(connection, msg, derr.message);
 		dbus_error_free(&derr);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 
 	match = g_slist_find_custom(devices, path, device_path_cmp);
 	if (!match)
-		return err_does_not_exist(connection, msg);
+		return error_device_does_not_exist(connection, msg);
 
 	reply = dbus_message_new_method_return(msg);
 	if (!reply)
@@ -979,7 +1008,7 @@ static DBusHandlerResult am_change_default_device(DBusConnection *conn,
 						DBUS_TYPE_STRING, &device->path,
 						DBUS_TYPE_INVALID);
 	else
-		return err_does_not_exist(connection, msg);
+		return error_device_does_not_exist(connection, msg);
 
 	default_dev = device;
 	device_store(device, TRUE);
@@ -1024,6 +1053,7 @@ static DBusSignalVTable manager_signals[] = {
 
 static void parse_stored_devices(char *key, char *value, void *data)
 {
+	bdaddr_t *src = data;
 	struct device *device;
 	bdaddr_t dst;
 
@@ -1031,7 +1061,7 @@ static void parse_stored_devices(char *key, char *value, void *data)
 		return;
 
 	str2ba(key, &dst);
-	device = manager_find_device(&dst, PKT_ROLE_AUTO, FALSE);
+	device = manager_find_device(&dst, NULL, FALSE);
 
 	if (device)
 		return;
@@ -1042,10 +1072,15 @@ static void parse_stored_devices(char *key, char *value, void *data)
 	if (!device)
 		return;
 
-	if (strstr(value, "headset"))
+	/* Change storage to source adapter */
+	bacpy(&device->store, src);
+
+	if (enabled.headset && strstr(value, "headset"))
 		device->headset = headset_init(device, NULL, 0);
-	if (strstr(value, "sink"))
+	if (enabled.sink && strstr(value, "sink"))
 		device->sink = sink_init(device);
+	if (enabled.control && strstr(value, "control"))
+		device->control = control_init(device);
 	add_device(device, FALSE);
 }
 
@@ -1074,10 +1109,9 @@ static void register_devices_stored(const char *adapter)
 
 	bacpy(&default_src, BDADDR_ANY);
 	dev_id = hci_get_route(&default_src);
-	if (dev_id < 0)
+	if (dev_id < 0 || hci_devba(dev_id, &default_src) < 0)
 		return;
 
-	hci_devba(dev_id, &default_src);
 	if (bacmp(&default_src, &src) != 0)
 		return;
 
@@ -1086,7 +1120,7 @@ static void register_devices_stored(const char *adapter)
 		return;
 
 	str2ba(addr, &dst);
-	device = manager_find_device(&dst, PKT_ROLE_AUTO, FALSE);
+	device = manager_find_device(&dst, NULL, FALSE);
 
 	if (device) {
 		info("Setting %s as default device", addr);
@@ -1192,7 +1226,7 @@ static int hsp_ag_record(sdp_buf_t *buf, uint8_t ch)
 	return ret;
 }
 
-static int hfp_ag_record(sdp_buf_t *buf, uint8_t ch)
+static int hfp_ag_record(sdp_buf_t *buf, uint8_t ch, uint32_t feat)
 {
 	sdp_list_t *svclass_id, *pfseq, *apseq, *root;
 	uuid_t root_uuid, svclass_uuid, ga_svclass_uuid;
@@ -1200,9 +1234,9 @@ static int hfp_ag_record(sdp_buf_t *buf, uint8_t ch)
 	sdp_profile_desc_t profile;
 	sdp_list_t *aproto, *proto[2];
 	sdp_record_t record;
-	uint16_t u16 = 0x0009;
 	sdp_data_t *channel, *features;
 	uint8_t netid = 0x01;
+	uint16_t sdpfeat;
 	sdp_data_t *network = sdp_data_alloc(SDP_UINT8, &netid);
 	int ret;
 
@@ -1233,7 +1267,8 @@ static int hfp_ag_record(sdp_buf_t *buf, uint8_t ch)
 	proto[1] = sdp_list_append(proto[1], channel);
 	apseq = sdp_list_append(apseq, proto[1]);
 
-	features = sdp_data_alloc(SDP_UINT16, &u16);
+	sdpfeat = (uint16_t) feat & 0xF;
+	features = sdp_data_alloc(SDP_UINT16, &sdpfeat);
 	sdp_attr_add(&record, SDP_ATTR_SUPPORTED_FEATURES, features);
 
 	aproto = sdp_list_append(0, apseq);
@@ -1352,10 +1387,10 @@ static void auth_cb(DBusPendingCall *call, void *data)
 	DBusError err;
 	const char *uuid;
 
-	if (headset_get_type(device) == SVC_HEADSET)
-		uuid = HSP_AG_UUID;
-	else
+	if (get_hfp_active(device))
 		uuid = HFP_AG_UUID;
+	else
+		uuid = HSP_AG_UUID;
 
 	dbus_error_init(&err);
 	if (dbus_set_error_from_message(&err, reply)) {
@@ -1388,7 +1423,7 @@ static gboolean ag_io_cb(GIOChannel *chan, GIOCondition cond, void *data)
 	socklen_t size;
 	const char *uuid;
 	struct device *device;
-	headset_type_t type;
+	gboolean hfp_active;
 
 	if (cond & G_IO_NVAL)
 		return FALSE;
@@ -1410,10 +1445,10 @@ static gboolean ag_io_cb(GIOChannel *chan, GIOCondition cond, void *data)
 	}
 
 	if (chan == hs_server) {
-		type = SVC_HEADSET;
+		hfp_active = FALSE;
 		uuid = HSP_AG_UUID;
 	} else {
-		type = SVC_HANDSFREE;
+		hfp_active = TRUE;
 		uuid = HFP_AG_UUID;
 	}
 
@@ -1435,7 +1470,7 @@ static gboolean ag_io_cb(GIOChannel *chan, GIOCondition cond, void *data)
 		return TRUE;
 	}
 
-	headset_set_type(device, type);
+	set_hfp_active(device, hfp_active);
 
 	if (!manager_authorize(&device->dst, uuid, auth_cb, device, NULL))
 		goto failed;
@@ -1501,12 +1536,15 @@ static GIOChannel *server_socket(uint8_t *channel)
 	return io;
 }
 
-static int headset_server_init(DBusConnection *conn, gboolean no_hfp)
+static int headset_server_init(DBusConnection *conn, GKeyFile *config)
 {
 	uint8_t chan = DEFAULT_HS_AG_CHANNEL;
 	sdp_buf_t buf;
+	gboolean no_hfp = FALSE;
+	GError *err = NULL;
+	uint32_t features;
 
-	if (!(enabled->headset || enabled->gateway))
+	if (!(enabled.headset || enabled.gateway))
 		return 0;
 
 	hs_server = server_socket(&chan);
@@ -1528,7 +1566,17 @@ static int headset_server_init(DBusConnection *conn, gboolean no_hfp)
 	}
 
 	g_io_add_watch(hs_server, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-						(GIOFunc) ag_io_cb, NULL);
+				(GIOFunc) ag_io_cb, NULL);
+
+	if (config) {
+		no_hfp = g_key_file_get_boolean(config, "Headset", "DisableHFP",
+						&err);
+		if (err) {
+			debug("audio.conf: %s", err->message);
+			g_error_free(err);
+			err = NULL;
+		}
+	}
 
 	if (no_hfp)
 		return 0;
@@ -1539,7 +1587,9 @@ static int headset_server_init(DBusConnection *conn, gboolean no_hfp)
 	if (!hf_server)
 		return -1;
 
-	if (hfp_ag_record(&buf, chan) < 0) {
+	features = headset_config_init(config);
+
+	if (hfp_ag_record(&buf, chan, features) < 0) {
 		error("Unable to allocate new service record");
 		return -1;
 	}
@@ -1554,7 +1604,7 @@ static int headset_server_init(DBusConnection *conn, gboolean no_hfp)
 	}
 
 	g_io_add_watch(hf_server, G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-						(GIOFunc) ag_io_cb, NULL);
+			(GIOFunc) ag_io_cb, NULL);
 
 	return 0;
 }
@@ -1582,14 +1632,54 @@ static void server_exit(void)
 	}
 }
 
-int audio_init(DBusConnection *conn, struct enabled_interfaces *enable,
-		gboolean no_hfp, gboolean sco_hci, int source_count)
+int audio_init(DBusConnection *conn, GKeyFile *config)
 {
-	int sinks, sources;
+	char *str;
+	GError *err = NULL;
 
 	connection = dbus_connection_ref(conn);
 
-	enabled = enable;
+	if (config) {
+		str = g_key_file_get_string(config, "General", "Enable", &err);
+
+		if (err) {
+			debug("audio.conf: %s", err->message);
+			g_error_free(err);
+			err = NULL;
+		} else {
+			if (strstr(str, "Headset"))
+				enabled.headset = TRUE;
+			if (strstr(str, "Gateway"))
+				enabled.gateway = TRUE;
+			if (strstr(str, "Sink"))
+				enabled.sink = TRUE;
+			if (strstr(str, "Source"))
+				enabled.source = TRUE;
+			if (strstr(str, "Control"))
+				enabled.control = TRUE;
+			g_free(str);
+		}
+
+		str = g_key_file_get_string(config, "General", "Disable", &err);
+
+		if (err) {
+			debug("audio.conf: %s", err->message);
+			g_error_free(err);
+			err = NULL;
+		} else {
+			if (strstr(str, "Headset"))
+				enabled.headset = FALSE;
+			if (strstr(str, "Gateway"))
+				enabled.gateway = FALSE;
+			if (strstr(str, "Sink"))
+				enabled.sink = FALSE;
+			if (strstr(str, "Source"))
+				enabled.source = FALSE;
+			if (strstr(str, "Control"))
+				enabled.control = FALSE;
+			g_free(str);
+		}
+	}
 
 	if (!dbus_connection_create_object_path(conn, AUDIO_MANAGER_PATH,
 						NULL, manager_unregister)) {
@@ -1597,20 +1687,17 @@ int audio_init(DBusConnection *conn, struct enabled_interfaces *enable,
 		goto failed;
 	}
 
-	if (headset_server_init(conn, no_hfp) < 0)
-		goto failed;
+	if (enabled.headset) {
+		if (headset_server_init(conn, config) < 0)
+			goto failed;
+	}
 
-	if (enable->sink)
-		sources = source_count;
-	else
-		sources = 0;
+	if (enabled.source || enabled.sink) {
+		if (a2dp_init(conn, config) < 0)
+			goto failed;
+	}
 
-	if (enable->source)
-		sinks = 1;
-	else
-		sinks = 0;
-
-	if (a2dp_init(conn, sources, sinks) < 0)
+	if (enabled.control && avrcp_init(conn) < 0)
 		goto failed;
 
 	if (!dbus_connection_register_interface(conn, AUDIO_MANAGER_PATH,
@@ -1750,15 +1837,19 @@ struct device *manager_find_device(bdaddr_t *bda, const char *interface,
 			continue;
 
 		if (interface && !strcmp(AUDIO_HEADSET_INTERFACE, interface)
-			&& !dev->headset)
+				&& !dev->headset)
 			continue;
 
 		if (interface && !strcmp(AUDIO_SINK_INTERFACE, interface)
-			&& !dev->sink)
+				&& !dev->sink)
 			continue;
 
 		if (interface && !strcmp(AUDIO_SOURCE_INTERFACE, interface)
-			&& !dev->source)
+				&& !dev->source)
+			continue;
+
+		if (interface && !strcmp(AUDIO_CONTROL_INTERFACE, interface)
+				&& !dev->control)
 			continue;
 
 		if (connected && !device_is_connected(dev, interface))
